#include <sys/poll.h>
#include <time.h>
#include <sys/syslog.h>
#include <sys/stat.h>
#include <signal.h>
#include "headers.h"
#include "network.h"
#include "uart.h"
#include "mc0203_protocol.h"

#define COMMAND_SET_CONFIG_RESPONSE 0x81
#define COMMAND_GET_CONFIG_RESPONSE 0x80

#define COMMAND_SET_CONFIG 0x01
#define COMMAND_GET_CONFIG 0x00

#define DELIMITER_FLAG 0x7e

char *INTERFACE_NAME = "eth0";
char *UART_NAME = "/dev/ttyAMA0";
char *PID_FILE_NAME = "/var/log/mc0203_daemon.pid";
int BAUDRATE = 115200;

const char *LOG_FILE_NAME = "/var/log/mc0203.log";
const off_t MAX_LOG_FILE_SIZE = 20 * 1024 * 1024; // 20MB
int LOG_FD = 0;


// returns 0 on success -1 on error
int become_daemon();

int create_pid_file(char *filename);

void read_cmd_args(int argc, char *argv[]);

//
int check_daemon();

void write_log(int type, char *message, int size);

int exec(struct uart_t *self);

void sigintHandler(int sig_num) {
    char log[256];
    snprintf(log, 256, "            DAEMON ENDED - errno = %d - SIGNAL = %d", errno, sig_num);
    write_log(LOG_DEBUG, log, strlen(log));
    close(LOG_FD);
    remove(PID_FILE_NAME);
    exit(0);
}

int main(int argc, char *argv[]) {
    char log_message[256];

    read_cmd_args(argc, argv);
    printf("start daemon\n");
    /* if (check_daemon()) { */
    /*     snprintf(log_message, 256, "DAEMON IS ALREADY RUNNING"); */
    /*     write_log(LOG_DEBUG, log_message, strlen(log_message)); */
    /*     exit(0); */
    /* } */
    /* if (become_daemon() != 0) { */
    /*     return EXIT_FAILURE; */
    /* } */
    /* create_pid_file(PID_FILE_NAME); */
    /* signal(SIGTERM, sigintHandler); */

    write_log(LOG_ERR, "           APP STARTED", strlen("           APP STARTED"));
    sprintf(log_message, "Network interface - %s", INTERFACE_NAME);
    write_log(LOG_ERR, log_message, strlen(log_message));
    sprintf(log_message, "UART - %s", UART_NAME);
    write_log(LOG_ERR, log_message, strlen(log_message));
    sprintf(log_message, "Baudrate - %d", BAUDRATE);
    write_log(LOG_ERR, log_message, strlen(log_message));


    struct uart_t main_uart;
    memset(&main_uart, 0, sizeof(struct uart_t));
    main_uart.invoke = exec;

    if (init_uart(&main_uart, UART_NAME, BAUDRATE) == -1) {
        sprintf(log_message, "%s: %s", "UART ERROR, app crashed", strerror(errno));
        write_log(LOG_ERR, log_message, strlen(log_message));
        sigintHandler(0);
        exit(1);
    }

    int d = init_network_interface(INTERFACE_NAME);
    if (d != 0) {
        sprintf(log_message, "%s: %s", "NETWORK ERROR, app crashed", strerror(errno));
        write_log(LOG_ERR, log_message, strlen(log_message));
        sigintHandler(0);
        exit(1);
    }

    struct pollfd uart_poll;
    uart_poll.fd = main_uart.fd;
    uart_poll.events = POLLIN;
    const int TIMEOUT = 100000;  // Timeout in milliseconds



    while (1) {
      printf("read cicle\n");
        int ret = poll(&uart_poll, 1, TIMEOUT);
        if (ret == -1) {
            perror("Failed to poll UART device");
            break;
        } else if (ret == 0) {
            // Timeout occurred
            continue;
        }
        if (uart_poll.revents & POLLIN) {
            read_uart_circle(&main_uart, DELIMITER_FLAG);
        }
    }


}

void read_cmd_args(int argc, char *argv[]) {
    char log[256];

    for (int i = 1; i < argc; i++) {
        if (strcmp(argv[i], "-b") == 0 && i + 1 < argc) {
            BAUDRATE = atoi(argv[++i]);
        } else {
            if (strcmp(argv[i], "-u") == 0 && i + 1 < argc) {
                UART_NAME = argv[++i];
            } else {
                if (strcmp(argv[i], "-n") == 0 && i + 1 < argc) {
                    INTERFACE_NAME = argv[++i];
                } else if (strcmp(argv[i], "-l") == 0 && i + 1 < argc) {
                    LOG_FILE_NAME = argv[++i];
                } else {
                    if (strcmp(argv[i], "--help") == 0) {
                        snprintf(log, 256, " help for %s:\n%s\n%s\n%s\n%s\n",
                                 argv[0],
                                 "-n - specify network interface name",
                                 "-u - specify uart interface name",
                                 "-b - specify uart baudrate",
                                 "-l - specify log file name");
                        printf(log);
                        exit(0);
                    } else {
                        snprintf(log, 256, "%s: invalid option -- '%s'\nTry '%s --help' for more information.\n",
                                 argv[0],
                                 argv[i],
                                 argv[0]);
                        write_log(LOG_ERR, log, strlen(log));
                        printf(log);
                        exit(0);
                    }
                }
            };

        }
    }
}

int exec(struct uart_t *self) {
    init_network_interface(INTERFACE_NAME);
    char index = 0xad;
    struct config_t config;
    unsigned char data_buffer[256], log_buffer[256];
    memset(data_buffer, 0, 200);

    int data_size = 0;
    int command;
    command = extract_data_from_command(self->message, 256, data_buffer, &data_size);
    switch (command) {
        case COMMAND_SET_CONFIG : {
            printf("\n\nCOMMAND_SET_CONFIG\n\n");
            convert_from_string(data_buffer, &(config.dhcp), &(config.ip), &(config.netmask), &(config.gw),
                                &(config.dns));
            snprintf(log_buffer, 256, "SET CONFIG - %s", data_buffer);
            write_log(LOG_DEBUG, log_buffer, strlen(log_buffer));
            char flag = set_c(&(config.dhcp), &(config.ip), &(config.netmask), &(config.gw),
                              &(config.dns), INTERFACE_NAME);
            if (flag != 0) {
                sprintf(data_buffer, "%s: %s", "SET_CONFIG ERROR - ", strerror(errno));
                write_log(LOG_ERR, data_buffer, strlen(data_buffer));

            }
            get_response(data_buffer, &data_size, &flag, 0, COMMAND_SET_CONFIG_RESPONSE, index);
            for (int i = 0; i < data_size; ++i) {
                printf("%03d ", data_buffer[i]);
            }
            write(self->fd, data_buffer, data_size);
            break;
        }
        case COMMAND_GET_CONFIG: {
            memset(self->message, 0, 200);
            printf("\n\nCOMMAND_GET_CONFIG\n\n");
            get_c(&config.dhcp, &(config.ip), &(config.netmask), &(config.gw), &(config.dns));
            convert_to_string(self->message, &(config.dhcp), &(config.ip), &(config.netmask), &(config.gw),
                              &(config.dns));

            get_response(data_buffer, &data_size, self->message, 199, COMMAND_GET_CONFIG_RESPONSE, index);
            printf("\n DATASIZE - %d \n RESPONSE FOR GET - ", data_size);
            for (int i = 0; i < data_size; ++i) {
                printf("%03d ", data_buffer[i]);
            }
            write(self->fd, data_buffer, data_size);
            break;
        }
    }
    return 0;
}

int become_daemon() {
    switch (fork()) {
        case -1:
            return -1;
        case 0:
            break;
        default:
            _exit(EXIT_SUCCESS);
    }

    if (setsid() == -1)
        return -1;

    switch (fork()) {
        case -1:
            return -1;
        case 0:
            break;
        default:
            _exit(EXIT_SUCCESS);
    }

    close(STDIN_FILENO);

    int fd = open("/dev/null", O_RDWR);
    if (fd != STDIN_FILENO)
        return -1;
    if (dup2(STDIN_FILENO, STDOUT_FILENO) != STDOUT_FILENO)
        return -2;
    if (dup2(STDIN_FILENO, STDERR_FILENO) != STDERR_FILENO)
        return -3;
    return 0;
}

int check_daemon() {
    int qwerty = open(PID_FILE_NAME, O_RDONLY);
    if (qwerty != -1) {
        close(qwerty);
        return 1;
    } else {
        close(qwerty);
        return 0;
    }
}

int create_pid_file(char *filename) {
    int qwerty = open(filename, O_CREAT | O_WRONLY);
    char buff[10];
    snprintf(buff, 10, "%d", getpid());
    write(qwerty, buff, strlen(buff));
    close(qwerty);
}

void write_log(int type, char *message, int size) {
    // Определение размера файла
    off_t file_size = 0;
    if (LOG_FD == 0) {
        LOG_FD = open(LOG_FILE_NAME, O_RDWR | O_APPEND);
    }
    if (LOG_FD != -1) {
        file_size = lseek(LOG_FD, 0, SEEK_END);
    }

    // Проверка размера файла
    if (file_size >= MAX_LOG_FILE_SIZE) {
        close(LOG_FD);
        LOG_FD = open(LOG_FILE_NAME, O_WRONLY | O_TRUNC | O_CREAT | O_APPEND);
        if (LOG_FD == -1) {
            perror("Failed to open log file");
            return;
        }
    } else {
        LOG_FD = open(LOG_FILE_NAME, O_WRONLY | O_CREAT | O_APPEND);
        if (LOG_FD == -1) {
            perror("Failed to open log file");
            return;
        }
    }

    time_t current_time = time(NULL);
    if (current_time == -1) {
        perror("Failed to get current time");
        close(LOG_FD);
        return;
    }

    // Перевод времени в строку
    struct tm *local_time = localtime(&current_time);
    if (local_time == NULL) {
        perror("Failed to convert time");
        close(LOG_FD);
        return;
    }
    char timestamp[20];
    strftime(timestamp, sizeof(timestamp), "%Y-%m-%d %H:%M:%S", local_time);

    // Добавление времени к логу
    char log_message[size + sizeof(timestamp) + 2]; // Extra 2 for ": " and '\0'
    snprintf(log_message, sizeof(log_message), "%s: %s", timestamp, message);

    // Запись лога
    write(LOG_FD, log_message, strlen(log_message));
    write(LOG_FD, "\n", 1);
}

